/**
 * Copyright 1993-2014 NVIDIA Corporation.  All rights reserved.
 *
 * Please refer to the NVIDIA end user license agreement (EULA) associated
 * with this source code for terms and conditions that govern your use of
 * this software. Any use, reproduction, disclosure, or distribution of
 * this software and related documentation outside the terms of the EULA
 * is strictly prohibited.
 *
 */

/**
 * Matrix multiplication (CUDA Kernel) on the device: C = A * B
 */


/*----------------------------------Includes-------------------------------------- */

// System includes
#include <stdio.h>
#include <assert.h>

// CUDA runtime
#include <cuda_runtime.h>

// Helper functions and utilities to work with CUDA
#include "helper_functions.h"

// include some convenience functions (names begin with "ue_" to distinguish them from CUDA)
#include "utility_environment.h"

// OpenGL utilities
#include "opengl_includes.h"




/*----------------------------------Global varaibles declaration-------------------------------------- */

int g_imgWidth = 512;                  // height and width of the image
int g_imgHeight = 512;

dim3 dimsA;
dim3 dimsB;

int block_length = 4;                //block length for square block considered (i.e. 4 X 4)

float *h_A = 0;						 
float *h_B = 0;                      // host buffers for matrix A, B & C resp.
float *h_C = 0;

float *d_A = 0;
float *d_B = 0;                      //GPU buffers for matrix A, B & C resp.
float *d_C = 0;


CPUBitmap *g_bitmap = 0;            //pointer to bit map object 


/*----------------------------------Methods-------------------------------------- */

/*kernel for the Matrix Multiplication
 Matrix multiplication (CUDA Kernel) on the device: C = A * B
 * wA is A's width and wB is B's width
 */
/*To be modified by the student for minimum operation*/
template <int BLOCK_SIZE> __global__ void
matrixMulCUDA(float *C, float *A, float *B, int wA, int wB)
{
    // Block index
    int bx = blockIdx.x;
    int by = blockIdx.y;

    // Thread index
    int tx = threadIdx.x;
    int ty = threadIdx.y;

	//global Indexes
	int idx = tx + bx * BLOCK_SIZE;
	int idy = ty + by * BLOCK_SIZE;

    // Index of the first sub-matrix of A processed by the block
    int aBegin = wA * BLOCK_SIZE * by;

    // Index of the last sub-matrix of A processed by the block
    int aEnd   = aBegin + wA - 1;

    // Step size used to iterate through the sub-matrices of A
    int aStep  = BLOCK_SIZE;

    // Index of the first sub-matrix of B processed by the block
    int bBegin = BLOCK_SIZE * bx;

    // Step size used to iterate through the sub-matrices of B
    int bStep  = BLOCK_SIZE * wB;

    // Csub is used to store the element of the block sub-matrix
    // that is computed by the thread
    float Csub = 0;

    // Loop over all the sub-matrices of A and B
    // required to compute the block sub-matrix
    for (int a = aBegin, b = bBegin;
         a <= aEnd;
         a += aStep, b += bStep)
    {

        // Declaration of the shared memory array As used to
        // store the sub-matrix of A
        __shared__ float As[BLOCK_SIZE][BLOCK_SIZE];

        // Declaration of the shared memory array Bs used to
        // store the sub-matrix of B
        __shared__ float Bs[BLOCK_SIZE][BLOCK_SIZE];

        // Load the matrices from device memory
        // to shared memory; each thread loads
        // one element of each matrix
        As[ty][tx] = A[a + wA * ty + tx];
        Bs[ty][tx] = B[b + wB * ty + tx];

        // Synchronize to make sure the matrices are loaded
        __syncthreads();

        // Multiply the two matrices together;
        // each thread computes one element
        // of the block sub-matrix
#pragma unroll
		
        for (int k = 0; k < BLOCK_SIZE; ++k)
        {
            Csub += As[ty][k] * Bs[k][tx];
        }
		

        // Synchronize to make sure that the preceding
        // computation is done before loading two new
        // sub-matrices of A and B in the next iteration
        __syncthreads();
    }

    // Write the block sub-matrix to device memory;
    // each thread writes one element
    int c = wB * BLOCK_SIZE * by + BLOCK_SIZE * bx;
    C[c + wB * ty + tx] = Csub;
}



/* initialize a data with random values from 0 to 10 */
void randomInit(float *data, int size)
{
	float LO = 0, HI = 20;
    for (int i = 0; i < size; ++i)
    {
        data[i] = LO + static_cast <float> (rand()) /( static_cast <float> (RAND_MAX/(HI-LO)));
    }
}

/* initialize data with a constant value "val" */
void constantInit(float *data, int size, float val)
{
	
    for (int i = 0; i < size; ++i)
    {
        data[i] = val;
    }
}

/* ideal function for glut */
void idleFunc( void )
{
	glutPostRedisplay();
}

/* glut display method for displaying the matrix*/
void display( void ) 
{
	 // Setup execution parameters
    dim3 threads(block_length, block_length);
	dim3 grid(dimsB.x / threads.x, dimsA.y / threads.y);
	
	matrixMulCUDA<4><<< grid, threads >>>(d_C, d_A, d_B, dimsA.x, dimsB.x);
		
	cudaDeviceSynchronize();  // wait until GPU is finished

	// copy GPU result back to host memory
    cudaMemcpy(h_C, d_C, dimsB.x*dimsA.y*sizeof(float), cudaMemcpyDeviceToHost);

	//convert the matrix to bitmap object and display the bitmap
	ue_convertMatrix2Bitmap(h_C, g_imgWidth, g_imgHeight, dimsB.x, dimsA.y, g_bitmap);
	g_bitmap->draw();
	glFlush();
}


/* Clean up memory */
void cleanUp()
{
    free(h_A);
    free(h_B);
    free(h_C);
    cudaFree(d_A);
    cudaFree(d_B);
    cudaFree(d_C);
}

/* static method used for glut callbacks */
void keyboard(unsigned char key, int x, int y)
{
	switch (key) 
	{
	case 27:              // keycode for the Escape key
		cleanUp();
		exit(0);
		break;
	}

	glutPostRedisplay();
}


/*----------------------------------Main Program-------------------------------------- */
int main(int argc, char **argv)
{

    printf("[Matrix Multiply Using CUDA] - Starting...\n");

    if (checkCmdLineFlag(argc, (const char **)argv, "help") ||
        checkCmdLineFlag(argc, (const char **)argv, "?"))
    {
        printf("Usage -device=n (n >= 0 for deviceID)\n");
        printf("      -wA=WidthA -hA=HeightA (Width x Height of Matrix A)\n");
        printf("      -wB=WidthB -hB=HeightB (Width x Height of Matrix B)\n");
        printf("  Note: Outer matrix dimensions of A & B matrices must be equal.\n");

        exit(EXIT_SUCCESS);
    }


	/////////////////////////////////////////////////////
	// setting and checking properties of the device  //
	////////////////////////////////////////////////////
    // By default, we use device 0, otherwise we override the device ID based on what is provided at the command line
    int devID = 0;

    if (checkCmdLineFlag(argc, (const char **)argv, "device"))
    {
        devID = getCmdLineArgumentInt(argc, (const char **)argv, "device");
        cudaSetDevice(devID);
    }

    cudaError_t error;
    cudaDeviceProp deviceProp;
    error = cudaGetDevice(&devID);

    if (error != cudaSuccess)
    {
        printf("cudaGetDevice returned error code %d, line(%d)\n", error, __LINE__);
    }

    error = cudaGetDeviceProperties(&deviceProp, devID);

    if (deviceProp.computeMode == cudaComputeModeProhibited)
    {
        fprintf(stderr, "Error: device is running in <Compute Mode Prohibited>, no threads can use ::cudaSetDevice().\n");
        exit(EXIT_SUCCESS);
    }

    if (error != cudaSuccess)
    {
        printf("cudaGetDeviceProperties returned error code %d, line(%d)\n", error, __LINE__);
    }
    else
    {
        printf("GPU Device %d: \"%s\" with compute capability %d.%d\n\n", devID, deviceProp.name, deviceProp.major, deviceProp.minor);
    }


	/* ########################################################################## */

	/////////////////////////////////////////////////////
	// setting the width and height of Matrices A & B  //
	/////////////////////////////////////////////////////

	//  length and width of the matrix A & B
    dimsA.x = 2*block_length;
	dimsA.y = 2*block_length;
	dimsB.x = 2*block_length;
	dimsB.y = 2*block_length;

    // width of Matrix A
    if (checkCmdLineFlag(argc, (const char **)argv, "wA"))
    {
        dimsA.x = getCmdLineArgumentInt(argc, (const char **)argv, "wA");
    }

    // height of Matrix A
    if (checkCmdLineFlag(argc, (const char **)argv, "hA"))
    {
        dimsA.y = getCmdLineArgumentInt(argc, (const char **)argv, "hA");
    }

    // width of Matrix B
    if (checkCmdLineFlag(argc, (const char **)argv, "wB"))
    {
        dimsB.x = getCmdLineArgumentInt(argc, (const char **)argv, "wB");
    }

    // height of Matrix B
    if (checkCmdLineFlag(argc, (const char **)argv, "hB"))
    {
        dimsB.y = getCmdLineArgumentInt(argc, (const char **)argv, "hB");
    }

    if (dimsA.x != dimsB.y)
    {
        printf("Error: outer matrix dimensions must be equal. (%d != %d)\n",
               dimsA.x, dimsB.y);
        exit(EXIT_FAILURE);
    }

    printf("MatrixA(%d,%d), MatrixB(%d,%d)\n", dimsA.x, dimsA.y, dimsB.x, dimsB.y);

	/* ########################################################################## */

	///////////////////////////////////////////////////////////////////////////
	// initialize GLUT, open display window, and register callback functions //
	///////////////////////////////////////////////////////////////////////////
	glutInit( &argc, argv );
	glutInitDisplayMode( GLUT_SINGLE | GLUT_RGBA );
	glutInitWindowSize( g_imgWidth, g_imgHeight);
	glutCreateWindow( "Matrix Visualization" );
	glutKeyboardFunc(keyboard);
	glutDisplayFunc(display);
	glutIdleFunc(idleFunc);

	/* ########################################################################## */

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Allocate host and device memories for matrices A, B & C  and initialize host memories for matrices A & B //
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////

    // Allocate host memory for matrices A and B
    unsigned int size_A = dimsA.x * dimsA.y;
    unsigned int mem_size_A = sizeof(float) * size_A;
    h_A = (float *)malloc(mem_size_A);
    unsigned int size_B = dimsB.x * dimsB.y;
    unsigned int mem_size_B = sizeof(float) * size_B;
    h_B = (float *)malloc(mem_size_B);

    // Initialize host memory
    const float valB = 1.0f/dimsA.y;

	//randomly initialize the matrix A 
    randomInit(h_A, size_A);
    constantInit(h_B, size_B, valB);



    // Allocate host matrix C
    dim3 dimsC(dimsB.x, dimsA.y, 1);
    unsigned int mem_size_C = dimsC.x * dimsC.y * sizeof(float);
    h_C = (float *) malloc(mem_size_C);

    if (h_C == NULL)
    {
        fprintf(stderr, "Failed to allocate host matrix C!\n");
        exit(EXIT_FAILURE);
    }


    error = cudaMalloc((void **) &d_A, mem_size_A);

    if (error != cudaSuccess)
    {
        printf("cudaMalloc d_A returned error code %d, line(%d)\n", error, __LINE__);
        exit(EXIT_FAILURE);
    }

    error = cudaMalloc((void **) &d_B, mem_size_B);

    if (error != cudaSuccess)
    {
        printf("cudaMalloc d_B returned error code %d, line(%d)\n", error, __LINE__);
        exit(EXIT_FAILURE);
    }

    error = cudaMalloc((void **) &d_C, mem_size_C);

    if (error != cudaSuccess)
    {
        printf("cudaMalloc d_C returned error code %d, line(%d)\n", error, __LINE__);
        exit(EXIT_FAILURE);
    }

    // copy host memory to device
    error = cudaMemcpy(d_A, h_A, mem_size_A, cudaMemcpyHostToDevice);

    if (error != cudaSuccess)
    {
        printf("cudaMemcpy (d_A,h_A) returned error code %d, line(%d)\n", error, __LINE__);
        exit(EXIT_FAILURE);
    }

    error = cudaMemcpy(d_B, h_B, mem_size_B, cudaMemcpyHostToDevice);

    if (error != cudaSuccess)
    {
        printf("cudaMemcpy (d_B,h_B) returned error code %d, line(%d)\n", error, __LINE__);
        exit(EXIT_FAILURE);
    }

	/* ########################################################################## */

	//////////////////////////////////////
	// display the resulting matrix C  //
	/////////////////////////////////////

	g_bitmap = new CPUBitmap(g_imgWidth, g_imgHeight);
   
	glutMainLoop();
    // cudaDeviceReset causes the driver to clean up all state. While
    // not mandatory in normal operation, it is good practice.  It is also
    // needed to ensure correct operation when the application is being
    // profiled. Calling cudaDeviceReset causes all profile data to be
    // flushed before the application exits
    cudaDeviceReset();

	/* ########################################################################## */

	return(0);

}
