﻿/**
 * Copyright 1993-2014 NVIDIA Corporation.  All rights reserved.
 *
 * Please refer to the NVIDIA end user license agreement (EULA) associated
 * with this source code for terms and conditions that govern your use of
 * this software. Any use, reproduction, disclosure, or distribution of
 * this software and related documentation outside the terms of the EULA
 * is strictly prohibited.
 *
 */


// System includes
#include <stdio.h>
#include <assert.h>
#include "helper_functions.h"

// CUDA runtime
#include <cuda_runtime.h>

// Helper functions and utilities to work with CUDA
#include <device_launch_parameters.h> // cuda keywords
#include <cuda.h>


/**
 * Matrix vector multiplication (CUDA Kernel) on the device: C = A * B
 * wA is A's width and wB is B's width
 */

//** TODO by the student **//
template <int BLOCK_SIZE> __global__ void
matrixVectorMulCUDA(float *Y, float *A, float *X, int wA)

{


}

void constantInit(float *data, int size, float val)
{
    for (int i = 0; i < size; ++i)
    {
        data[i] = val;
    }
}

/**
 * Run a simple test of matrix vector multiplication using CUDA
 */
int matrixVectorMultiply(int argc, char **argv, int block_size, dim3 &dimsA, dim3 &dimsX)
{
    // Allocate host memory for matriX A and vector X
    unsigned int size_A = dimsA.x * dimsA.y;
    unsigned int mem_size_A = sizeof(float) * size_A;
    float *h_A = (float *)malloc(mem_size_A);
    unsigned int size_X = dimsX.x;
    unsigned int mem_size_X = sizeof(float) * size_X;
    float *h_X = (float *)malloc(mem_size_X);

    // Initialize host memory
    const float valX = 0.01f;
    constantInit(h_A, size_A, 1.0f);
    constantInit(h_X, size_X, valX);

    // Allocate device memory
    float *d_A, *d_X, *d_Y;

    // Allocate host vector Y (result of the product A * X)
    dim3 dimsY(dimsA.x, 1 , 1);
    unsigned int mem_size_Y = dimsY.x * sizeof(float);
    float *h_Y = (float *) malloc(mem_size_Y);

    if (h_Y == NULL)
    {
        fprintf(stderr, "Failed to allocate host vector X!\n");
        exit(EXIT_FAILURE);
    }

    cudaError_t error;

    error = cudaMalloc((void **) &d_A, mem_size_A);

    if (error != cudaSuccess)
    {
        printf("cudaMalloc d_A returned error code %d, line(%d)\n", error, __LINE__);
        exit(EXIT_FAILURE);
    }

    error = cudaMalloc((void **) &d_X, mem_size_X);

    if (error != cudaSuccess)
    {
        printf("cudaMalloc d_X returned error code %d, line(%d)\n", error, __LINE__);
        exit(EXIT_FAILURE);
    }

    error = cudaMalloc((void **) &d_Y, mem_size_Y);

    if (error != cudaSuccess)
    {
        printf("cudaMalloc d_Y returned error code %d, line(%d)\n", error, __LINE__);
        exit(EXIT_FAILURE);
    }

    // copy host memory to device
    error = cudaMemcpy(d_A, h_A, mem_size_A, cudaMemcpyHostToDevice);

    if (error != cudaSuccess)
    {
        printf("cudaMemcpy (d_A,h_A) returned error code %d, line(%d)\n", error, __LINE__);
        exit(EXIT_FAILURE);
    }

    error = cudaMemcpy(d_X, h_X, mem_size_X, cudaMemcpyHostToDevice);

    if (error != cudaSuccess)
    {
        printf("cudaMemcpy (d_X,h_X) returned error code %d, line(%d)\n", error, __LINE__);
        exit(EXIT_FAILURE);
    }

    // Setup execution parameters
    dim3 threads(block_size);
    dim3 grid(dimsX.x / threads.x);

    // Create and start timer
    printf("Computing result using CUDA Kernel...\n");

    // Performs warmup operation using matrixMul CUDA kernel
    if (block_size == 256)
    {
        matrixVectorMulCUDA<256><<< grid, threads >>>(d_Y, d_A, d_X, dimsA.x);
    }
    else
    {
        matrixVectorMulCUDA<1024><<< grid, threads >>>(d_Y, d_A, d_X, dimsA.x);
    }

    printf("done\n");

    cudaDeviceSynchronize();

	// Copy result from device to host
    error = cudaMemcpy(h_Y, d_Y, mem_size_Y, cudaMemcpyDeviceToHost);

    // Allocate CUDA events that we'll use for timing
    cudaEvent_t start;
    error = cudaEventCreate(&start);

    if (error != cudaSuccess)
    {
        fprintf(stderr, "Failed to create start event (error code %s)!\n", cudaGetErrorString(error));
        exit(EXIT_FAILURE);
    }

    cudaEvent_t stop;
    error = cudaEventCreate(&stop);

    if (error != cudaSuccess)
    {
        fprintf(stderr, "Failed to create stop event (error code %s)!\n", cudaGetErrorString(error));
        exit(EXIT_FAILURE);
    }

    // Record the start event
    error = cudaEventRecord(start, NULL);

    if (error != cudaSuccess)
    {
        fprintf(stderr, "Failed to record start event (error code %s)!\n", cudaGetErrorString(error));
        exit(EXIT_FAILURE);
    }

	
    // Execute the kernel
    int nIter = 300;

    for (int j = 0; j < nIter; j++)
    {
        if (block_size == 256)
        {
            matrixVectorMulCUDA<256><<< grid, threads >>>(d_Y, d_A, d_X, dimsA.x);
        }
        else
        {
            matrixVectorMulCUDA<1024><<< grid, threads >>>(d_Y, d_A, d_X, dimsA.x);
        }
    }

    // Record the stop event
    error = cudaEventRecord(stop, NULL);

    if (error != cudaSuccess)
    {
        fprintf(stderr, "Failed to record stop event (error code %s)!\n", cudaGetErrorString(error));
        exit(EXIT_FAILURE);
    }

    // Wait for the stop event to complete
    error = cudaEventSynchronize(stop);

    if (error != cudaSuccess)
    {
        fprintf(stderr, "Failed to synchronize on the stop event (error code %s)!\n", cudaGetErrorString(error));
        exit(EXIT_FAILURE);
    }

    float msecTotal = 0.0f;
    error = cudaEventElapsedTime(&msecTotal, start, stop);

    if (error != cudaSuccess)
    {
        fprintf(stderr, "Failed to get time elapsed between events (error code %s)!\n", cudaGetErrorString(error));
        exit(EXIT_FAILURE);
    }

    // Compute and print the performance
    float msecPerMatrixMul = msecTotal / nIter;
    double flopsPerMatrixVectorMul = 2.0 * (double)dimsA.x * (double)dimsX.x;
    double gigaFlops = (flopsPerMatrixVectorMul * 1.0e-9f) / (msecPerMatrixMul / 1000.0f);
    printf(
        "Performance= %.2f GFlop/s, Time= %.3f msec, Size= %.0f Ops, WorkgroupSize= %u threads/block\n",
        gigaFlops,
        msecPerMatrixMul,
        flopsPerMatrixVectorMul,
        threads.x);

    // Copy result from device to host
    error = cudaMemcpy(h_Y, d_Y, mem_size_Y, cudaMemcpyDeviceToHost);

    if (error != cudaSuccess)
    {
        printf("cudaMemcpy (h_C,d_C) returned error code %d, line(%d)\n", error, __LINE__);
        exit(EXIT_FAILURE);
    }

    printf("Checking computed result for correctness: ");
    bool correct = true;

    // test relative error by the formula
    //     |<x, y>_cpu - <x,y>_gpu|/<|x|, |y|>  < eps
    double eps = 1.e-6 ; // machine zero

    for (int i = 0; i < (int)(dimsY.x); i++)
    {
        double abs_err = fabs(h_Y[i] - (dimsA.x * valX));
        double dot_length = dimsA.x;
        double abs_val = fabs(h_Y[i]);
        double rel_err = abs_err/abs_val/dot_length ;

        if (rel_err > eps)
        {
            printf("Error! Matrix[%05d]=%.8f, ref=%.8f error term is > %E\n", i, h_Y[i], dimsA.x*valX, eps);
            correct = false;
        }
    }

    printf("%s\n", correct ? "Result = PASS" : "Result = FAIL");

    // Clean up memory
    free(h_A);
    free(h_X);
    free(h_Y);
    cudaFree(d_A);
    cudaFree(d_X);
    cudaFree(d_Y);


    // cudaDeviceReset causes the driver to clean up all state. While
    // not mandatory in normal operation, it is good practice.  It is also
    // needed to ensure correct operation when the application is being
    // profiled. Calling cudaDeviceReset causes all profile data to be
    // flushed before the application exits
    cudaDeviceReset();

    if (correct)
    {
        return EXIT_SUCCESS;
    }
    else
    {
        return EXIT_FAILURE;
    }
}


/**
 * Program main
 */
int main(int argc, char **argv)
{
    printf("[Matrix Vector Multiply Using CUDA] - Starting...\n");

    if (checkCmdLineFlag(argc, (const char **)argv, "help") ||
        checkCmdLineFlag(argc, (const char **)argv, "?"))
    {
        printf("Usage -device=n (n >= 0 for deviceID)\n");
        printf("      -wA=WidthA -hA=HeightA (Width x Height of Matrix A)\n");
        printf("      -lX=LengthX (Length of vector X)\n");
        printf("  Note: Width of Matrix A = length of vector X\n");

        exit(EXIT_SUCCESS);
    }

    // By default, we use device 0, otherwise we override the device ID based on what is provided at the command line
    int devID = 0;

    if (checkCmdLineFlag(argc, (const char **)argv, "device"))
    {
        devID = getCmdLineArgumentInt(argc, (const char **)argv, "device");
        cudaSetDevice(devID);
    }


	//check the device errors
    cudaError_t error;
    cudaDeviceProp deviceProp;
    error = cudaGetDevice(&devID);

    if (error != cudaSuccess)
    {
        printf("cudaGetDevice returned error code %d, line(%d)\n", error, __LINE__);
    }

    error = cudaGetDeviceProperties(&deviceProp, devID);

    if (deviceProp.computeMode == cudaComputeModeProhibited)
    {
        fprintf(stderr, "Error: device is running in <Compute Mode Prohibited>, no threads can use ::cudaSetDevice().\n");
        exit(EXIT_SUCCESS);
    }

    if (error != cudaSuccess)
    {
        printf("cudaGetDeviceProperties returned error code %d, line(%d)\n", error, __LINE__);
    }
    else
    {
        printf("GPU Device %d: \"%s\" with compute capability %d.%d\n\n", devID, deviceProp.name, deviceProp.major, deviceProp.minor);
    }



    // Use a larger block size for Fermi and above
    int block_size = (deviceProp.major < 2) ? 256 : 1024;

	//assign the default dimensions of the matrix A
	dim3 dimsA(2*block_size, 50*block_size, 1);
    dim3 dimsX(2*block_size, 1, 1);

    // width of Matrix A
    if (checkCmdLineFlag(argc, (const char **)argv, "wA"))
    {
        dimsA.x = getCmdLineArgumentInt(argc, (const char **)argv, "wA");
    }

    // height of Matrix A
    if (checkCmdLineFlag(argc, (const char **)argv, "hA"))
    {
        dimsA.y = getCmdLineArgumentInt(argc, (const char **)argv, "hA");
    }

    // width of Matrix B
    if (checkCmdLineFlag(argc, (const char **)argv, "lX"))
    {
        dimsX.x = getCmdLineArgumentInt(argc, (const char **)argv, "wB");
    }


    if (dimsA.x != dimsX.x)
    {
        printf("Error: Width of Matrix A  and length of vector X are not equal (%d != %d)\n",
               dimsA.x, dimsX.x);
        exit(EXIT_FAILURE);
    }

    printf("MatrixA(%d,%d), VectorX(%d)\n", dimsA.x, dimsA.y, dimsX.x);

    int matrix_vector_multiply_result = matrixVectorMultiply(argc, argv, block_size, dimsA, dimsX);

    exit(matrix_vector_multiply_result);
}
